;; Generated package description from scad-mode.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "scad-mode" "98.0.0.20260330.71707" "A major mode for editing OpenSCAD code" '((emacs "29.1") (compat "30")) :commit "44481331536516e34940315f67fa58f99f62a589" :maintainer '(("Len Trigg" . "lenbok@gmail.com") ("Daniel Mendler" . "mail@daniel-mendler.de")) :keywords '("languages") :url "https://github.com/openscad/emacs-scad-mode")
